(*1*)

let extrait_liste_hasard l =
  let n = List.length !l in
  let rec aux accu fin_liste i =
    match fin_liste with
        x::ffl ->
          if i=0
          then (l := accu@ffl; x)
          else aux (x::accu) ffl (i-1)
  in aux [] !l (Random.int n)
;;

(*2*)

type labyrinthe = {longueur : int;
                   hauteur : int;
                   mur_vert : bool array array;
                   mur_hori : bool array array};;

type coord = {x : int; y : int};;

let origine = {x=0; y=0};;

type mur = {pos : coord; vertical : bool};;

let initialise_labyrinthe p q = {longueur = p;
                                 hauteur = q;
                                 mur_vert = Array.make_matrix (p-1) q true;
                                 mur_hori = Array.make_matrix p (q-1) true}
;;

(*3*)

let cree_labyrinthe longueur hauteur =
  (* initialisation des variables locales *)
  let connecte = Array.make_matrix longueur hauteur false in
  let frontiere = ref [{pos={x=0;y=0};vertical=true};
                       {pos={x=0;y=0};vertical=false}] in
  let nbre_cases_non_connectees = ref (longueur*hauteur) in
  let laby = initialise_labyrinthe longueur hauteur in
    (* fin initialisation *)
  let case_adjacente_non_connectee m =
    let a = m.pos.x in
    let b = m.pos.y in
      if not (connecte.(a).(b)) then (true, {x=a; y=b})
      else if m.vertical && not (connecte.(a+1).(b)) then (true, {x=a+1; y=b})
      else if not (m.vertical) && not (connecte.(a).(b+1)) then (true, {x=a; y=b+1})
       else (false, {x=0; y=0}) in
  let abattre_mur m =
    if m.vertical then laby.mur_vert.(m.pos.x).(m.pos.y) <- false
    else laby.mur_hori.(m.pos.x).(m.pos.y) <- false in
  let ajoute_murs_adjacents case =
    let a = case.x in
    let b = case.y in
      if a>0 && not(connecte.(a-1).(b)) then frontiere:={pos={x=a-1;y=b};vertical=true}::(!frontiere);
      if a<longueur-1 && not(connecte.(a+1).(b)) then frontiere:={pos={x=a;y=b};vertical=true}::(!frontiere);
      if b>0 && not(connecte.(a).(b-1)) then frontiere:={pos={x=a;y=b-1};vertical=false}::(!frontiere);
      if b<hauteur-1 && not(connecte.(a).(b+1)) then frontiere:={pos={x=a;y=b};vertical=false}::(!frontiere);
      () in
    while !nbre_cases_non_connectees > 0 do
      let m = extrait_liste_hasard frontiere in
      let (existe, c) = case_adjacente_non_connectee m in
        if existe
        then begin
          abattre_mur m;
          connecte.(c.x).(c.y) <- true;
          ajoute_murs_adjacents c;
          decr nbre_cases_non_connectees
        end
    done;
    laby
;;

(*8*)

#load "graphics.cma";;
open Graphics;;
open_graph " 800x600";;

(*9*)

let taille_case laby =
  min ((size_x () - 10) / laby.longueur) ((size_y () - 10) / laby.hauteur);;

let coo_labyrinthe laby i j =
  let tc = taille_case laby in
    (i*tc + 5, j*tc + 5)
;;

let vaa laby i j =
  let (a, b) = coo_labyrinthe laby i j in
    moveto a b
;;

let lignevers laby i j =
  let (a, b) = coo_labyrinthe laby i j in
    lineto a b
;;


(*10*)

let trace_mur laby m =
  vaa laby (m.pos.x+1) (m.pos.y+1);
  if m.vertical then lignevers laby (m.pos.x+1) m.pos.y
  else lignevers laby m.pos.x (m.pos.y+1)
;;

let trace_labyrinthe laby =
  set_line_width 3;
  vaa laby 0 0;
  lignevers laby 0 laby.hauteur;
  lignevers laby laby.longueur laby.hauteur;
  lignevers laby laby.longueur 0;
  lignevers laby 0 0;
  set_line_width 1;
  for i = 0 to (laby.longueur - 2) do
    for j = 0 to (laby.hauteur - 1) do
      if laby.mur_vert.(i).(j) then trace_mur laby {pos={x=i;y=j};vertical=true}
    done;
  done;
  for i = 0 to (laby.longueur - 1) do
    for j = 0 to (laby.hauteur - 2) do
      if laby.mur_hori.(i).(j) then trace_mur laby {pos={x=i;y=j};vertical=false}
    done;
  done
;;

let coordonnees_centre_case laby i j =
  let (a, b) = coo_labyrinthe laby i j in
  let tdc = (taille_case laby) / 2 in
    (a+tdc, b+tdc)
;;

let positionne_centre_case laby i j =
  let (x, y) = coordonnees_centre_case laby i j in
    moveto x y
;;

let imprime_centre_case laby chaine i j =
  positionne_centre_case laby i j;
  let (a, b) = text_size chaine in
    rmoveto (-a/2) (-b/2);
    draw_string chaine
;;


let laby = cree_labyrinthe 50 50;;

trace_labyrinthe (laby);;


(* RETROUVER LE CHEMIN *)

type direction = Inconnue | Haut | Bas | Gauche | Droite | Arrivee;;

let construit_carte laby =
  let carte = Array.make_matrix laby.longueur laby.hauteur Inconnue in
  let rec construit_carte_depuis i j =
    match (i, j) with
      | (0, 0) -> ()
      | _ ->
          if i > 0 && carte.(i-1).(j) = Inconnue && not (laby.mur_vert.(i-1).(j))
          then
            begin
              carte.(i-1).(j) <- Droite;
              construit_carte_depuis (i-1) j
            end;
          if i < laby.longueur-1 && carte.(i+1).(j) = Inconnue && not (laby.mur_vert.(i).(j))
          then
            begin
              carte.(i+1).(j) <- Gauche;
              construit_carte_depuis (i+1) j
            end;
          if j>0 && carte.(i).(j-1) = Inconnue && not (laby.mur_hori.(i).(j-1))
          then
            begin
              carte.(i).(j-1) <- Haut;
              construit_carte_depuis i (j-1)
            end;
          if j < laby.hauteur-1 && carte.(i).(j+1) = Inconnue && not (laby.mur_hori.(i).(j))
          then
            begin
              carte.(i).(j+1) <- Bas;
              construit_carte_depuis i (j+1)
            end
  in
    carte.(laby.longueur-1).(laby.hauteur-1) <- Arrivee;
    construit_carte_depuis (laby.longueur-1) (laby.hauteur-1);
    carte
;;

let carte = construit_carte laby;;
    
let reconstitue_chemin laby carte =
  let rec chemine itineraire =
    match itineraire with
      | [] -> failwith "Pas possible"
      | (i, j)::fl -> match carte.(i).(j) with
          | Arrivee -> List.rev itineraire
          | Gauche -> chemine ((i-1, j)::itineraire)
          | Droite -> chemine ((i+1, j)::itineraire)
          | Haut -> chemine ((i, j+1)::itineraire)
          | Bas -> chemine ((i, j-1)::itineraire)
          | Inconnue -> failwith "Pas possible..."
  in chemine [(0, 0)]
;;

let ch = reconstitue_chemin laby carte;;


let trace_chemin laby ch =
  let rec etape = function
    | [] -> ()
    | (i, j)::fch ->
        let (x, y) = coordonnees_centre_case laby i j in
          lineto x y;
          etape fch
  in
    match ch with
      | [] -> ()
      | (i, j)::sch ->
          set_color red;
          positionne_centre_case laby i j;
          etape sch;
          set_color black
;;


trace_chemin laby ch;;




